﻿using Eingabeverwaltung.Metamodell;
using Eingabeverwaltung.Parser.ADT_Parser;
using System;
using Antlr4.Runtime.Tree;
using Antlr4.Runtime;
using Antlr4.Runtime.Misc;
using System.Collections.Generic;
using System.Linq;

namespace Eingabeverwaltung.Parser
{
    public class ADT_Listener : IADTListener
    {
        // object on which will be build
        private Container container;

        // helping objects for building
        private List<Sort> op_sorts;

        private Listener_Helper LH;

        public ADT_Listener(Container container, Listener_Helper LH)
        {
            this.container = container;
            op_sorts = null;

            this.LH = LH;
        }

        public void EnterAdt([NotNull] ADTParser.AdtContext context) { }
        public void ExitAdt([NotNull] ADTParser.AdtContext context)
        {
            Coords coords = new Coords(context);
            if (LH.adt_sorts == null) LH.errors.Add(new SemanticAnalysisError(coords,"adt_sorts = null at ExitAdt"));
            if (LH.adt_ops == null) LH.errors.Add(new SemanticAnalysisError(coords,"adt_ops = null at ExitAdt"));
            if (LH.adt_axioms == null) LH.adt_axioms = new List<Axiom>();
            if (LH.adt_sorts.Count < 1) LH.errors.Add(new SemanticAnalysisError(coords, "adt_sorts is empty at ExitAdt"));
            if (LH.adt_ops.Count < 1) LH.errors.Add(new SemanticAnalysisError(coords, "adt_ops is empty at ExitAdt"));
            container.ADT = new ADT(LH.adt_sorts, LH.adt_ops, LH.adt_axioms);
        }
        public void EnterName([NotNull] ADTParser.NameContext context) { }
        public void ExitName([NotNull] ADTParser.NameContext context) { }

        public void EnterSorts([NotNull] ADTParser.SortsContext context) => LH.adt_sorts = new List<Sort>();
        public void ExitSorts([NotNull] ADTParser.SortsContext context)
        {
            if (LH.adt_sorts.Count < 1) LH.errors.Add(new SemanticAnalysisError(new Coords(context),"adt_sorts is empty at ExitSorts"));
        }

        public void EnterSort([NotNull] ADTParser.SortContext context) { }
        public void ExitSort([NotNull] ADTParser.SortContext context)
        {
            string sortName = context.BEZ().GetText();
            Coords coords = new Coords(context);
            if (context.Parent.GetChild(0).GetText() == "sorts")
            {
                Test_Sorts_all_unique(sortName,coords);
                LH.adt_sorts.Add(new Sort(sortName));
                return;
            }
            if (new List<string>(){"operations","constructors"}.Contains(context.Parent.Parent.GetChild(0).GetText()))
            {
                Sort sort = LH.findSortInADT_Sorts(sortName);
                Test_Operation_all_Sorts_were_defined_in_ADT(sortName, coords);
                op_sorts.Add(sort);
                return;
            }
            if (context.Parent.Parent.GetChild(0).GetText() == "vars") return; // this job is done in ExitVar
            throw new SemanticAnalysisDeadlyError(coords,"ExitSort with Context = " + context.GetText() + " in parent "+ context.Parent.GetChild(0).GetText() + " was not able to find the sense of its life.");
        }
        public void EnterConstructors([NotNull] ADTParser.ConstructorsContext context) => LH.adt_ops = new List<Operation>();
        public void ExitConstructors([NotNull] ADTParser.ConstructorsContext context)
        {
            Coords coords = new Coords(context);
            if (LH.adt_ops.Count < 1) LH.errors.Add(new SemanticAnalysisError(coords, "adt_ops is empty at ExitConstructors"));
            // check for each sort: numer & type of constructors is ok
            foreach (Sort sort in LH.adt_sorts)
                if (sort.GetBuildingConstructors().Count > 0 && sort.GetBasicConstructors().Count == 0)
                    LH.errors.Add(new SemanticAnalysisError(coords,"The sort "+sort.name+" has building constructors, but no basic constructors."));
        }

        public void EnterOperations([NotNull] ADTParser.OperationsContext context) 
        {
            if (LH.adt_ops == null) LH.adt_ops = new List<Operation>(); 
        }
        public void ExitOperations([NotNull] ADTParser.OperationsContext context)
        {
            if (LH.adt_ops.Count < 1) LH.errors.Add(new SemanticAnalysisError(new Coords(context), "adt_ops is empty at ExitOperations"));
        }

        public void EnterOperation(ADTParser.OperationContext context) => op_sorts = new List<Sort>();
        public void ExitOperation(ADTParser.OperationContext context)
        {
            Coords coords = new Coords(context);
            if (op_sorts.Count < 1) LH.errors.Add(new SemanticAnalysisError(coords, "op_sorts is empty at ExitOperation"));
            string operationName = context.BEZ().GetText();
            Test_Operations_all_unique(operationName,new Coords(context));
            Operation operation;
            if (context.Parent.GetChild(0).GetText() == "operations")
                operation = new Operation(operationName, op_sorts);
            else if (context.Parent.GetChild(0).GetText() == "constructors")
            {
                bool isBasic = true;
                for (int s = 0; s < op_sorts.Count - 1; s++)
                    if (op_sorts[s] == op_sorts.Last()) { isBasic = false; break; }
                if (isBasic) operation = new BasicConstructor(operationName, op_sorts);
                else operation = new BuildingConstructor(operationName, op_sorts);
                op_sorts.Last().SetConstructor((Constructor)operation);
            }
            else throw new SemanticAnalysisDeadlyError(coords,"ExitOperation with Context = " + context.GetText() + " in parent " + context.Parent.GetChild(0).GetText() + " was not able to find the sense of its life.");
            LH.adt_ops.Add(operation);
        }

        public void EnterVars([NotNull] ADTParser.VarsContext context) => LH.EnterVars();
        public void ExitVars([NotNull] ADTParser.VarsContext context) { }

        public void EnterVar([NotNull] ADTParser.VarContext context) { }
        public void ExitVar([NotNull] ADTParser.VarContext context)
            => LH.ExitVar(context.BEZ().GetText(), context.sort().BEZ().GetText(),new Coords(context));

        public void EnterAxioms([NotNull] ADTParser.AxiomsContext context) => LH.adt_axioms = new List<Axiom>();
        public void ExitAxioms([NotNull] ADTParser.AxiomsContext context)
        {
            if (LH.adt_axioms.Count < 1) LH.errors.Add(new SemanticAnalysisError(new Coords(context),"adt_axioms is empty at ExitAxioms"));
        }

        public void EnterAxiom([NotNull] ADTParser.AxiomContext context) => LH.trees = new List<Tree>();
        public void ExitAxiom([NotNull] ADTParser.AxiomContext context)
        {
            string axiomName = context.BEZ().GetText();
            Test_Axioms_all_unique(axiomName,new Coords(context));
            LH.ExitEq(new Coords(context));
            LH.adt_axioms.Add(new Axiom(axiomName, LH.equation));
        }

        // Trees
        public void EnterMytree([NotNull] ADTParser.MytreeContext context) { }
        public void ExitMytree([NotNull] ADTParser.MytreeContext context) { }

        public void EnterEq_s([NotNull] ADTParser.Eq_sContext context) { }
        public void ExitEq_s([NotNull] ADTParser.Eq_sContext context) { }
        // classic Trees
        public void EnterClassicTree([NotNull] ADTParser.ClassicTreeContext context) { }
        public void ExitClassicTree([NotNull] ADTParser.ClassicTreeContext context) { }

        public void EnterTreeLeaf([NotNull] ADTParser.TreeLeafContext context) { }
        public void ExitTreeLeaf([NotNull] ADTParser.TreeLeafContext context)
            => LH.ExitTreeLeaf(context.BEZ().GetText(), new Coords(context));

        public void EnterTreeNode([NotNull] ADTParser.TreeNodeContext context) { }
        public void ExitTreeNode([NotNull] ADTParser.TreeNodeContext context)
            => LH.ExitTreeNode(context.BEZ().GetText(), context.ChildCount, context.GetText(), new Coords(context));
        // function Trees
        public void EnterFTree([NotNull] ADTParser.FTreeContext context) { }
        public void ExitFTree([NotNull] ADTParser.FTreeContext context) { }
        public void EnterFTreeRoot([NotNull] ADTParser.FTreeRootContext context) { }
        public void ExitFTreeRoot([NotNull] ADTParser.FTreeRootContext context)
        => LH.ExitFTreeRoot(new Coords(context));

        public void EnterFTreeLeft([NotNull] ADTParser.FTreeLeftContext context) { }
        public void ExitFTreeLeft([NotNull] ADTParser.FTreeLeftContext context) { }
        public void EnterFTreeRight([NotNull] ADTParser.FTreeRightContext context) { }
        public void ExitFTreeRight([NotNull] ADTParser.FTreeRightContext context) { }

        public void EnterFTreeBracket([NotNull] ADTParser.FTreeBracketContext context) { }
        public void ExitFTreeBracket([NotNull] ADTParser.FTreeBracketContext context) { }

        public void EnterFTreeBez([NotNull] ADTParser.FTreeBezContext context) { }

        public void ExitFTreeBez([NotNull] ADTParser.FTreeBezContext context)
            => LH.ExitFTreeBez(context.BEZ().GetText(), new Coords(context));

        // useless Interface requirements
        public void EnterEveryRule(ParserRuleContext ctx) { }
        public void ExitEveryRule(ParserRuleContext ctx) { }
        public void VisitErrorNode(IErrorNode node) { }
        public void VisitTerminal(ITerminalNode node) { }

        // Alle Offiziell notwendige Typtests

        /// <summary>
        /// This Test is activated always when a new Sort gets added to the sortList of the ADT,
        /// so as result all Sorts are unique.
        /// </summary>
        /// <param name="sortName"></param>
        private void Test_Sorts_all_unique(string sortName, Coords coords)
        {
            if (LH.findSortInADT_Sorts(sortName) != null) LH.errors.Add(new SemanticAnalysisError(coords,"The Sort " + sortName + " was defined more than 1 times in Sorts of ADT!"));
        }
        /// <summary>
        /// This Test is activated always when a new Operation gets added to the operationList of the ADT,
        /// so as result all Operations are unique.
        /// </summary>
        /// <param name="operationName"></param>
        private void Test_Operations_all_unique(string operationName, Coords coords)
        {
            if (LH.findOperationInADT_Ops(operationName,0,false,coords) != null) LH.errors.Add(new SemanticAnalysisError(coords, "The Operation " + operationName + " was defined more than 1 times in Operations of ADT!"));
        }
        /// <summary>
        /// This Test is activated always when a new Sort gets added to the sortList of an Operation,
        /// so as result every Sort of every Operation was defined in the ADT.
        /// </summary>
        /// <param name="sortName"></param>
        private void Test_Operation_all_Sorts_were_defined_in_ADT(string sortName, Coords coords)
        {
            if (LH.findSortInADT_Sorts(sortName) == null) LH.errors.Add(new SemanticAnalysisError(coords, "Found a Sort " + sortName + " in Operations that was not defined in Sorts!"));
        }
        /// <summary>
        /// This Test is activated always when a new Axiom gets added to the axiomList of the ADT,
        /// so as result all Axioms are unique. 
        /// </summary>
        /// <param name="axiomName"></param>
        private void Test_Axioms_all_unique(string axiomName, Coords coords)
        {
            foreach (Axiom axiom in LH.adt_axioms)
            {
                if (axiom.name == axiomName) LH.errors.Add(new SemanticAnalysisError(coords, "The Axiom " + axiomName + " was defined more than 1 times in Axioms of ADT!"));
            }
        }
        /// <summary>
        ///  This is done in LH.ExitEq() by Test_Equation_both_sides_have_same_Sort. 
        ///  This dummy Method is for knowing, that this Test is implemented.
        /// </summary>
        private void Test_Axiom_both_sides_have_same_Sort() {/* Nothing here need to be. */}
    }
}