﻿using Eingabeverwaltung.Metamodell;
using System;
using System.Collections.Generic;
using Ueberpruefung.CheckTransformation;
using Ueberpruefung.CheckInduction;
using Eingabeverwaltung;

namespace Ueberpruefung
{
    public static class Ueberpruefer
    {
        public static void ueberpruefe(Container con)
        {
            // pre checkings and ordering of lemmata, their dependencies, usefulness and pre required (in)validness for cycles
            analyseCyclicDependencies(con.Proof.lemmata);
            List<SingleTask> allTasks = new List<SingleTask>() { con.Task.task };
            allTasks.AddRange(con.Task.lemmaTasks);
            analyseUsefulness(allTasks,con.Proof);
            // main checkings of errors
            foreach (Lemma lemma in con.Proof.lemmata)
            {
                if (lemma.OwnTask != null) lemma.IsProved = true;
                else if (!lemma.accept(new CheckLemmaIsAcyclic()))                      // 20
                        lemma.IsProved = false;
            }
            con.accept(new CheckLemmaTaskIsUsed());                                     // 23
            // now check all proofs in their topological order
            // (lemmata are ordered by analyseCyclicDependencies, main proof is last in this order)
            foreach (Lemma lemma in con.Proof.lemmata)
            {
                bool isProved = true;
                isProved &= lemma.accept(new CheckTask_Satisfied());                    // 1, 2
                isProved &= lemma.accept(new CheckIndTask_Satisfied());                 // 38, 39
                isProved &= ueberpruefe(lemma.proof, lemma);
                if (lemma.IsProved == null) lemma.IsProved = isProved;
            }
            con.accept(new CheckTask_Satisfied());                                      // 1, 2
            con.accept(new CheckIndTask_Satisfied());                                   // 38, 39
            ueberpruefe(con.Proof.proof,null);
            // post checkings for assessment: usefulness of lemmata to induction cases of their tasks with induction config
            analyseUsefulnessToIndCases(allTasks,con.Proof);
        }

        private static bool ueberpruefe(SingleProof sp, Lemma Owner)
        {
            if (sp is Transformation tr) return ueberpruefe(tr, Owner);
            else if (sp is Induction ind) return ueberpruefe(ind, Owner);
            else throw new NotImplementedException("Ueberpruefer can only check transformation or induction as single proof");
        }

        private static bool ueberpruefe(Induction ind, Lemma Owner)
        {
            bool isProved = true;
            isProved &= ind.accept(new CheckIndVarNotFixed());                          // 30
            if (ind.accept(new CheckIndVarHasConstructors()))                           // 31
            {
                foreach (InductionCase ic in ind.indCases)
                {
                    isProved &= ic.accept(new CheckTask_Satisfied());                   // 1, 2
                    isProved &= ic.accept(new CheckIndPartEquation(ind));               // 34 (ic)
                    isProved &= ic.accept(new CheckIndPartEquationIsFixed(ind));        // 35 (ic)
                    isProved &= ic.accept(new CheckIndHypNotUsedInIA());                // 36
                    // this errors don't making lemmata unproved
                    ic.accept(new CheckIndPartEquationIsNotFixed(ind));                 // 40 (ic)
                    ic.accept(new CheckIndCaseType());                                  // 41
                }
                isProved &= ind.accept(new CheckIndCasesAllExisting());                 // 32
                isProved &= ind.accept(new CheckIndCaseTypeIsValid());                  // 33
                // check ihs (must be done after check ic: requires ic.buildingParameterVariable)                
                foreach (InductionHypothesis ih in ind.indHyps)
                {
                    bool IHisProved = true;
                    IHisProved &= ih.accept(new CheckIndPartEquation(ind));             // 34 (ih)
                    IHisProved &= ih.accept(new CheckIndPartEquationIsFixed(ind));      // 35 (ih)
                    ih.accept(new CheckIndPartEquationIsNotFixed(ind));                 // 40 (ih) (don't makes lemma or ih unproved)
                    // must be set before checkings of transformation (for error 37)
                    ih.IsProved = IHisProved;
                    isProved &= IHisProved;
                }
                ind.accept(new CheckIndHypsUnique());                                   // 42
                ind.accept(new CheckIndHypCanBeUseful());                               // 43
                // check transformations
                foreach (InductionCase ic in ind.indCases)
                    isProved &= ueberpruefe(ic.proof, Owner);
            }
            else isProved = false;
            return isProved;
        }

        private static bool ueberpruefe(Transformation tr, Lemma Owner) {
            bool isProved = true;
            foreach (TransformationStep ts in tr.transformationSteps)
            {
                bool formalError = false;
                foreach (Substitution sub in ts.substitutions)
                {
                    if (sub.accept(new CheckSubstitutionVariable_DefinedIn_ADT()))      // 3
                    {
                        if (!sub.accept(new CheckSubstitution_NoSortChange()))          // 4
                            formalError = true;
                    }
                    else formalError = true;
                }
                if (!ts.accept(new CheckTransformationStep_NoSortChange()))             // 5
                    formalError = true;
                isProved &= !formalError;
                if (!formalError)
                {
                    bool termPart_isSubTreeFrom_starTree =
                         ts.accept(new CheckTermPart_isSubTreeFrom_StartTree());        // 6
                    bool termPart_match_RuleStartTree =
                         ts.accept(new CheckTermPart_match_RuleStartTree());            // 7
                    bool termPart_match_RuleEndTree =
                         ts.accept(new CheckTermPart_match_RuleEndTree());              // 8
                    isProved &= termPart_isSubTreeFrom_starTree;
                    if (termPart_match_RuleStartTree && termPart_match_RuleEndTree 
                        && !ts.ruleStartTree().Equals(ts.ruleEndTree())) // when both sides are equal, direction can't be wrong
                    {
                        // assume wrong direction 
                        // 10 here don't marks the error,
                        // because termpart match also ruleStartTree
                        if (ts.accept(new CheckSubstitution_for_RuleEndTree()))                    // 10
                        {
                            // direction is wrong
                            ts.errors.Add(new Error(8));
                            if (termPart_isSubTreeFrom_starTree)
                                isProved &= ts.accept(new CheckResultTree_for_RuleEndTree());      // 12
                        }
                        else
                        {   // direction is correct
                            if (ts.accept(new CheckSubstitution_for_RuleStartTree())               // 9 
                                && termPart_isSubTreeFrom_starTree)
                                isProved &= ts.accept(new CheckResultTree_for_RuleStartTree());    // 11
                            else isProved = false;
                        }
                    }
                    else
                    {
                        if (termPart_match_RuleStartTree)
                        {
                            if (ts.accept(new CheckSubstitution_for_RuleStartTree())               // 9
                                && termPart_isSubTreeFrom_starTree)
                                isProved &= ts.accept(new CheckResultTree_for_RuleStartTree());    // 11
                            else isProved = false;
                        }
                        else
                        {
                            if (termPart_match_RuleEndTree)
                            {
                                ts.errors.Add(new Error(8));
                                if (ts.accept(new CheckSubstitution_for_RuleEndTree())             // 10
                                    && termPart_isSubTreeFrom_starTree)
                                    isProved &= ts.accept(new CheckResultTree_for_RuleEndTree());  // 12
                                else isProved = false;
                            }
                            else { ts.errors.Add(new Error(7)); isProved = false; }
                        }
                    }
                    ts.accept(new CheckSubstitutionVariable_isUsedInRule());                       // 13
                    isProved &= ts.accept(new CheckUsedLemmaIsProved(Owner));                      // 21
                    isProved &= ts.accept(new CheckFixedVarsNotSubstituted());                     // 37
                }
            }
            return isProved;
        }    
        
        /// <summary>
        /// This Method is helping the Ueberpruefer checking cycles in lemmata
        /// by analysing their dependencies.
        /// @pre:  All cyclicDependencies of lemmata are empty.
        /// @post: The whole cycle of the lemma ist saved in its cyclicDependencies.
        /// </summary>
        /// <param name="lemmata"></param>
        private static void analyseCyclicDependencies(List<Lemma> lemmata)
        {
            // 1. get all direct dependencies
            void writeDirectCyclicDependencies(Lemma lemma, Transformation tr) 
            {
                foreach (TransformationStep ts in tr.transformationSteps)
                {
                    if (ts.rule is Lemma usedLemma)
                    {
                        bool found = false;
                        foreach (Lemma current in lemmata)
                            if (current == usedLemma)
                            {
                                if (!lemma.cyclicDependencies.Contains(usedLemma)) lemma.cyclicDependencies.Add(usedLemma);
                                found = true; break;
                            }
                        if (!found) throw new KontextException(ts.rule.coords, "Error in Ueberpruefung: Could not find Lemma " + usedLemma.name + " in Lemmata!");
                    }
                }
            }
            foreach(Lemma lemma in lemmata)
            {
                if (lemma.proof is Transformation tr)
                    writeDirectCyclicDependencies(lemma, tr);
                else if (lemma.proof is Induction ind)
                    foreach (InductionCase indC in ind.indCases)
                        writeDirectCyclicDependencies(lemma, indC.proof);
                else throw new KontextException(lemma.proof.coords,"Found a proof in Lemma that is neither Transformation nor Induction!");
            }
            //Console.WriteLine("After 1."); print(lemmata,true);
            // 2. remove Lemmata not depending from cyclic Lemmata by topological sort
            List<Lemma> nonCyclic = new List<Lemma>();
            List<Lemma> remaining = new List<Lemma>(lemmata);
            List<Lemma> moved;
            do
            {
                moved = new List<Lemma>(); int i = -1;
                while (i < remaining.Count - 1) // move independent lemmata to nonCyclic
                {
                    i++;
                    if (remaining[i].cyclicDependencies.Count == 0)
                    {
                        nonCyclic.Add(remaining[i]);
                        moved.Add(remaining[i]);
                        remaining.RemoveAt(i); i--;
                    }
                }
                foreach (Lemma movedLemma in moved) // remove Dependencies from moved
                    foreach (Lemma remLemma in remaining) remLemma.cyclicDependencies.Remove(movedLemma);
            } while (moved.Count != 0);
            //Console.WriteLine("After 2."); print(lemmata,true);
            // 3. distinguish cyclic Lemmata and Lemmata depending from cyclic Lemmata:
            //    Only real cycle elements must be marked with that error
            List<Lemma> useCyclic = new List<Lemma>();
            Dictionary<Lemma, List<Lemma>> transitiveHullSaving = new Dictionary<Lemma, List<Lemma>>();
            List<Lemma> getTransitiveHull(Lemma current, List<Lemma> hull)
            {
                foreach(Lemma dep in current.cyclicDependencies)
                    if (!hull.Contains(dep))
                    {
                        hull.Add(dep);
                        if (transitiveHullSaving.ContainsKey(dep))
                        {
                            foreach (Lemma saved in transitiveHullSaving[dep])
                                if (!hull.Contains(saved)) hull.Add(saved);
                        }else getTransitiveHull(dep, hull);
                    }                        
                return hull;
            }
            for (int i = 0; i< remaining.Count; i++)
            {
                Lemma current = remaining[i];
                List<Lemma> transitiveHull = getTransitiveHull(current, new List<Lemma>());
                transitiveHullSaving.Add(current, transitiveHull);
                if (transitiveHull.Contains(current))
                {// found element of cycle
                    current.cyclicDependencies.Clear();
                    current.cyclicDependencies.AddRange(transitiveHull);
                    //current.cyclicDependencies.Remove(current);
                }
                else
                {// not an element of cycle
                    current.cyclicDependencies.Clear();
                    useCyclic.Add(current);
                    remaining.RemoveAt(i); i--;
                    foreach (Lemma remLemma in remaining) remLemma.cyclicDependencies.Remove(current);
                }
            }
            //Console.WriteLine("After 3."); print(lemmata,true);
            // 4. move the topological order to list "lemmata"
            lemmata.Clear();
            lemmata.AddRange(nonCyclic);
            lemmata.AddRange(remaining); // = cycle(s)
            lemmata.AddRange(useCyclic);
            // result is already contained in the list "lemmata"
        }

        /// <summary>
        /// This Method finds all Lemmata that are main proof for a task
        /// and let every Lemma know for which task it is useful.
        /// Also Checks Error 22.
        /// @pre: tasks[0].name = "main task"
        /// </summary>
        /// <param name="tasks"></param>
        /// <param name="proof"></param>
        private static void analyseUsefulness(List<SingleTask> tasks, Proof proof)
        {
            // 1. find all Lemmata that are main proof for a task
            List<Lemma> mainLemmata = new List<Lemma>();
            for (int i = 1; i < tasks.Count; i++)
            {
                SingleTask task = tasks[i];
                foreach (Lemma lemma in proof.lemmata)
                {
                    if (lemma.name == task.name && lemma.accept(new CheckLemmaEqualsTask(task)))   // 22
                    {
                        lemma.OwnTask = task;
                        lemma.usefulFor.Add(new TaskAssignment(task));
                        mainLemmata.Add(lemma);
                    }
                }
            }
            // 2. analyse usefulness of other lemmata            
            List<Lemma> usefulLemmata = GetAllUsefulLemmata(new List<Lemma>(), proof.proof);
            foreach (Lemma usefulLemma in usefulLemmata)
                usefulLemma.usefulFor.Add(new TaskAssignment(tasks[0]));
            foreach (Lemma mainLemma in mainLemmata)
            {
                usefulLemmata = GetAllUsefulLemmata(new List<Lemma>(), mainLemma.proof);
                foreach (Lemma usefulLemma in usefulLemmata)
                    usefulLemma.usefulFor.Add(new TaskAssignment(mainLemma.OwnTask));
            }
        }

        /// <summary>
        /// Analyses usefulness of lemmata to the induction cases of their tasks with induction config. <br></br>
        /// @pre: All induction cases must have analysed their constructor. <br></br>
        /// Does nothing if an induction case don't knows its constructor.
        /// </summary>
        /// <param name="tasks"></param>
        /// <param name="proof"></param>
        private static void analyseUsefulnessToIndCases(List<SingleTask> tasks, Proof proof)
        {
            // 1. get all lemmata inclusive a lemma for main proof
            Lemma mainProofLemma = new Lemma(proof.proof.coords,"main proof", tasks[0].eq, tasks[0].fixedVars, proof.proof);
            mainProofLemma.OwnTask = tasks[0];
            mainProofLemma.usefulFor.Add(new TaskAssignment(tasks[0]));
            List<Lemma> allLemmata = new List<Lemma>() { mainProofLemma };
            allLemmata.AddRange(proof.lemmata);
            List<Lemma> mainIndLemmata = new List<Lemma>();
            // 2. get all main lemma of tasks with induction config
            foreach (Lemma lemma in allLemmata)
                if (lemma.OwnTask != null && lemma.OwnTask.config is InductionConfig) mainIndLemmata.Add(lemma);
            // 3. specialize the usefulness to induction cases of the mainIndLemmata
            foreach(Lemma mainIndLemma in mainIndLemmata)
            {
                if (mainIndLemma.proof is Induction ind)
                {
                    foreach(InductionCase ic in ind.indCases)
                    {
                        if (ic.Constructor == null) continue; // those inductioncases don't get assessed, so it should not count to the usefulness
                        foreach (Lemma usefulLemma in GetAllUsefulLemmata(new List<Lemma>(), ic.proof))
                        {
                            TaskAssignment ta = null;
                            foreach(TaskAssignment currentTa in usefulLemma.usefulFor)
                                if (currentTa.task == mainIndLemma.OwnTask) { ta = currentTa; break; }
                            if (ta == null) throw new Exception("Could not find Task Assignment to task "+mainIndLemma.OwnTask.name+" for Lemma "+usefulLemma.name+" in its Task Assignments!"); // there are no useful coords
                            ta.AddIndCase(ic.Constructor);
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Finds all transitive useful lemmata to the given proof. <br></br>
        /// Start with input result = new List. <br></br>
        /// Adds all useful Lemmata to input list result and returns the same list.
        /// </summary>
        /// <param name="result"></param>
        /// <param name="proof"></param>
        /// <returns></returns>
        private static List<Lemma> GetAllUsefulLemmata(List<Lemma> result, SingleProof proof)
        {
            List<Lemma> GetAllUsefulLemmataTR(List<Lemma> result, List<Lemma> newLemmata, Transformation tr)
            {
                foreach (TransformationStep ts in tr.transformationSteps)
                {

                    if (ts.rule is Lemma lemma && !result.Contains(lemma) && lemma.OwnTask == null)
                    {
                        // lemma with error 22 should not be added, because it would be a main proof for a task
                        bool lemmaNotMainProof = true;
                        foreach (Error err in lemma.errors)
                            if (err.errorCode == 22) lemmaNotMainProof = false;
                        if (lemmaNotMainProof)
                        {
                            result.Add(lemma);
                            newLemmata.Add(lemma);
                        }
                    }
                }
                return result;
            }
            List<Lemma> newLemmata = new List<Lemma>();
            if (proof is Transformation tr)
                GetAllUsefulLemmataTR(result, newLemmata, tr);
            else if (proof is Induction ind)
            {
                foreach (InductionCase indC in ind.indCases)
                    GetAllUsefulLemmataTR(result, newLemmata, indC.proof);
            }
            else throw new NotImplementedException("Found a single proof that is neither Transformation nor Induction!");
            foreach (Lemma newLemma in newLemmata)
                GetAllUsefulLemmata(result, newLemma.proof);
            return result;
        }
    }
}
